from typing import Literal
from typing import Optional
from typing import overload

from gyver.attrs import call_init
from gyver.attrs import define

from gyver.utils import lazyfield

from .config import CacheConfig
from .interface import AsyncCacheInterface
from .mapper import AsyncCacheMap
from .mock import MockAsyncCache
from .redis import AsyncRedisWrapper


@define
class AsyncCacheProvider:
    config: CacheConfig
    test: bool

    @overload
    def __init__(self, *, config: CacheConfig):
        ...

    @overload
    def __init__(self, *, test: Literal[True]):
        ...

    def __init__(
        self, *, config: Optional[CacheConfig] = None, test: bool = False
    ) -> None:
        call_init(self, config, test)

    @staticmethod
    def _make_concrete(
        config: Optional[CacheConfig], test: bool
    ) -> AsyncCacheInterface:
        if test:
            return MockAsyncCache()
        return AsyncRedisWrapper(config) if config is not None else AsyncRedisWrapper()

    @lazyfield
    def interface(self):
        return self._make_concrete(self.config, self.test)

    def mapper(self, name: str) -> AsyncCacheMap:
        return AsyncCacheMap(self.interface, name)



import typing

import sqlalchemy as sa
from gyver.attrs import define
from gyver.attrs import info
from sqlalchemy.ext.asyncio import create_async_engine

from gyver.config import AdapterConfigFactory
from gyver.context import atomic
from gyver.database.context.asyncio import AsyncSessionAdapter
from gyver.database.context.sync import SessionAdapter
from gyver.database.typedef import Driver
from gyver.utils.lazy import lazyfield

from .config import DatabaseConfig
from .context import AsyncConnectionAdapter
from .context import SaAdapter
from .utils import make_uri

db_config_loader = AdapterConfigFactory().maker(DatabaseConfig, "db")


@define
class DatabaseAdapter:
    config: DatabaseConfig = info(default=db_config_loader)
    db_kwargs: dict[str, typing.Any] = info(default_factory=dict)

    def __post_init__(self) -> None:
        if self.config.driver is not Driver.SQLITE:
            self.db_kwargs.update(
                {
                    "pool_size": self.config.pool_size,
                    "pool_recycle": self.config.pool_recycle,
                    "max_overflow": self.config.max_overflow,
                }
            )

    @lazyfield
    def async_engine(self):
        return create_async_engine(make_uri(self.config), **self.db_kwargs)

    @lazyfield
    def engine(self):
        return sa.create_engine(make_uri(self.config, sync=True), **self.db_kwargs)

    def context(self):
        context = SaAdapter(engine=self.engine).context()
        return atomic(context) if self.config.autotransaction else context

    def async_context(self):
        context = AsyncConnectionAdapter(engine=self.async_engine).context()
        return atomic(context) if self.config.autotransaction else context

    def session(self):
        context = SessionAdapter(SaAdapter(engine=self.engine)).context()
        return atomic(context) if self.config.autotransaction else context

    def async_session(self):
        context = AsyncSessionAdapter(
            AsyncConnectionAdapter(engine=self.async_engine)
        ).context()
        return atomic(context) if self.config.autotransaction else context



from abc import ABC
from abc import abstractmethod

from gyver.attrs import define
from sqlalchemy.sql import Select

from gyver.database.typedef import ClauseType

from . import comp as cp
from .interface import ApplyClause
from .interface import Comparator
from .where import Where


@define
class Paginate(ApplyClause, ABC):
    type_ = ClauseType.APPLY
    limit: int
    offset: int

    @abstractmethod
    def apply(self, query: Select) -> Select:
        raise NotImplementedError

    @staticmethod
    def none():
        return _NullPaginate(limit=0, offset=0)

    def __bool__(self):
        return isinstance(self, _NullPaginate)


class LimitOffsetPaginate(Paginate):
    def apply(self, query: Select) -> Select:
        return query.limit(self.limit).offset(self.offset)


@define
class FieldPaginate(Paginate):
    field: str = "id"
    jump_comparison: Comparator = cp.greater

    def apply(self, query: Select) -> Select:
        return query.where(
            Where(self.field, self.offset, self.jump_comparison).bind(
                query.selected_columns  # type: ignore
            )
        ).limit(self.limit)


class _NullPaginate(Paginate):
    def apply(self, query: Select) -> Select:
        return query



from typing import Any
from typing import Generator
from typing import Sequence
from typing import Union

from gyver.attrs import define
from pydantic import BaseModel
from pydantic.fields import ModelField

from gyver.config.adapter.interface import FieldResolverStrategy
from gyver.config.config import MISSING


@define
class PydanticResolverStrategy(FieldResolverStrategy[ModelField]):
    field: ModelField

    def cast(self) -> type:
        return self.field.outer_type_

    def names(self) -> Sequence[str]:
        return (self.field.name, self.field.alias)

    def init_name(self) -> str:
        return self.field.name

    def default(self) -> Union[Any, type[MISSING]]:
        if self.field.default not in (None, Ellipsis):
            return self.field.default
        return (
            MISSING
            if self.field.default_factory is None
            else self.field.default_factory
        )

    @staticmethod
    def iterfield(
        config_class: type[BaseModel],
    ) -> Generator[ModelField, Any, Any]:
        yield from config_class.__fields__.values()



from collections import defaultdict
from typing import Mapping
from typing import Optional
from urllib.parse import parse_qs
from urllib.parse import quote

from gyver.attrs import mutable

from gyver.url.encode import Encodable


@mutable(eq=False)
class Query(Encodable):
    """
    Represents a query string and provides methods to manipulate and encode it.

    :param querystr: The input query string.
    """

    params: defaultdict[str, list[str]]

    def __init__(self, querystr: str) -> None:
        """
        Initializes the Query object with the provided query string.

        :param querystr: The input query string to parse.
        """
        self.params = defaultdict(list)
        self.params |= parse_qs(querystr, keep_blank_values=True)

    def encode(self) -> str:
        """
        Encodes the query object into a string representation.

        :return: The encoded query string.
        """
        return "&".join(
            "=".join((quote(key, safe=""), quote(item or "", safe="")))
            for key, value in self.params.items()
            for item in value
        )

    def omit_empty_equal(self) -> str:
        """
        Encodes the query object into a string representation, omitting the equal sign if a value is empty.

        :return: The encoded query string.
        """
        return "&".join(
            "=".join(map(quote, (key, item))) if item else quote(key)
            for key, value in self.params.items()
            for item in value
        )

    def add(
        self, args: Optional[Mapping[str, str]] = None, /, **params: str
    ) -> "Query":
        """
        Adds query parameters to the Query object.

        :param args: Additional query parameters as a mapping object.
        :param params: Additional query parameters as keyword arguments.
        :return: The modified Query object.
        """
        query = {**(args or {}), **params}
        for key, value in query.items():
            self.params[key].append(value)
        return self

    def set(
        self, args: Optional[Mapping[str, str]] = None, /, **params: str
    ) -> "Query":
        """
        Sets the query parameters, replacing any existing parameters.

        :param args: New query parameters as a mapping object.
        :param params: New query parameters as keyword arguments.
        :return: The modified Query object.
        """
        query = {**(args or {}), **params}
        self.params.clear()
        self.add(query)
        return self

    def __setitem__(self, key: str, value: str) -> None:
        """
        Sets a query parameter using the square bracket syntax.

        :param key: The parameter key.
        :param value: The parameter value.
        """
        self.add({key: value})

    def __getitem__(self, key: str) -> list[str]:
        """
        Retrieves the values of a query parameter using the square bracket syntax.

        :param key: The parameter key.
        :return: The list of parameter values.
        """
        return self.params[key]

    def remove(self, *keys: str) -> "Query":
        """
        Removes query parameters from the Query object.

        :param keys: The parameter keys to remove.
        :return: The modified Query object.
        """
        for key in keys:
            self.params.pop(key, None)
        return self

    def sort(self) -> "Query":
        """
        Sorts the query parameters in lexicographic order based on the parameter names.

        :return: The modified Query object.
        """
        self.params = defaultdict(list, sorted(self.params.items()))
        return self



from time import time
from typing import Any
from typing import Generic
from typing import TypeVar

from gyver.attrs import define
from gyver.attrs import mutable

from gyver.exc import InvalidParamType

T = TypeVar("T")


STARTTIME_ATTR = "_gyver_starttime_"


@mutable
class ResourceProxy:
    _target: Any
    _gyver_starttime_: float

    def __init__(self, target: Any, starttime: float):
        object.__setattr__(self, "_target", target)
        object.__setattr__(self, "_gyver_starttime_", starttime)

    def __getattr__(self, name: str) -> Any:
        return getattr(self._target, name)

    def __setattr__(self, name: str, value: Any) -> None:
        if name == "_gyver_starttime_":
            super().__setattr__(name, value)
        else:
            setattr(self._target, name, value)

    def __delattr__(self, name: str) -> None:
        delattr(self._target, name)

    @classmethod
    def as_any(cls, target: Any, starttime: float) -> Any:
        return cls(target, starttime)


@define
class Resource(Generic[T]):
    resource: T

    @classmethod
    def from_now(cls, resource: T) -> "Resource[T]":
        """
        Create a `Resource` object with the current timestamp as the `last_usage` value.

        :param resource: The underlying resource.
        :return: The `Resource` object.
        """
        current_ts = time()
        return cls(ResourceProxy.as_any(resource, current_ts))

    @classmethod
    def from_resource(cls, resource: T) -> "Resource[T]":
        """
        Create a `Resource` object from an existing resource.

        :param resource: The underlying resource.
        :return: The `Resource` object.
        :raises: InvalidParamType if the resource was not initialized correctly.
        """
        if not hasattr(resource, STARTTIME_ATTR):
            raise InvalidParamType(
                "Resource was not initialized correctly", resource
            ) from None
        return cls(resource)

    @property
    def last_usage(self) -> float:
        return getattr(self.resource, STARTTIME_ATTR)

    def get(self) -> T:
        """
        Get the underlying resource.

        :return: The underlying resource.
        """
        return self.resource



from typing import Any
from typing import Callable
from typing import Protocol
from typing import TypeVar
from typing import Union
from typing import overload


class MISSING:
    pass


def _default_cast(a: Any):
    return a


T = TypeVar("T")


class ConfigLike(Protocol):
    def get(
        self,
        name: str,
        cast: Callable = _default_cast,
        default: Union[Any, type[MISSING]] = MISSING,
    ) -> Any:
        ...

    @overload
    def __call__(
        self,
        name: str,
        cast: Union[Callable[[Any], T], type[T]] = _default_cast,
        default: type[MISSING] = MISSING,
    ) -> T:
        ...

    @overload
    def __call__(
        self,
        name: str,
        cast: Union[Callable[[Any], T], type[T]] = _default_cast,
        default: T = ...,
    ) -> T:
        ...

    def __call__(
        self,
        name: str,
        cast: Union[Callable[[Any], T], type[T]] = _default_cast,
        default: Union[T, type[MISSING]] = MISSING,
    ) -> T:
        ...



import os
from os import environ
from pathlib import Path
from typing import Any
from typing import Callable
from typing import Iterator
from typing import MutableMapping
from typing import TypeVar
from typing import Union
from typing import overload

from gyver.attrs import define
from gyver.attrs import info

from gyver.config.interface import MISSING
from gyver.config.interface import _default_cast
from gyver.exc import InvalidCast
from gyver.exc import MissingName
from gyver.utils.exc import panic
from gyver.utils.lazy import lazyfield

T = TypeVar("T")


@define
class EnvMapping(MutableMapping[str, str]):
    mapping: MutableMapping[str, str] = environ
    already_read: set[str] = info(default_factory=set)

    def __getitem__(self, name: str):
        val = self.mapping[name]
        self.already_read.add(name)
        return val

    def __setitem__(self, name: str, value: str):
        if name in self.already_read:
            raise panic(KeyError, f"{name} already read, cannot change its value")
        self.mapping[name] = value

    def __delitem__(self, name: str) -> None:
        if name in self.already_read:
            raise panic(KeyError, f"{name} already read, cannot delete")
        del self.mapping[name]

    def __iter__(self) -> Iterator[str]:
        yield from self.mapping

    def __len__(self) -> int:
        return len(self.mapping)


default_mapping = EnvMapping()


@define
class Config:
    env_file: Union[str, Path, None] = None
    mapping: EnvMapping = default_mapping

    def __post_init__(self):
        if self.env_file and os.path.isfile(self.env_file):
            self.file_values.update(dict(self._read_file(self.env_file)))

    @lazyfield
    def file_values(self):
        return {}

    def _read_file(self, env_file: Union[str, Path]):
        with open(env_file) as buf:
            for line in buf:
                if line.startswith("#"):
                    continue
                name, value = line.split("=", 1)
                yield name.strip(), value.strip()

    def _cast(self, name: str, val: Any, cast: Callable) -> Any:
        try:
            val = cast(val)
        except Exception as e:
            raise panic(InvalidCast, f"{name} received an invalid value {val}") from e
        else:
            return val

    def _get_val(
        self, name: str, default: Union[Any, type[MISSING]] = MISSING
    ) -> Union[Any, type[MISSING]]:
        return self.mapping.get(name, self.file_values.get(name, default))

    def get(
        self,
        name: str,
        cast: Callable = _default_cast,
        default: Union[Any, type[MISSING]] = MISSING,
    ) -> Any:
        val = self._get_val(name, default)
        if val is MISSING:
            raise panic(MissingName, f"{name} not found and no default was given")
        return self._cast(name, val, cast)

    @overload
    def __call__(
        self,
        name: str,
        cast: Union[Callable[[Any], T], type[T]] = _default_cast,
        default: type[MISSING] = MISSING,
    ) -> T:
        ...

    @overload
    def __call__(
        self,
        name: str,
        cast: Union[Callable[[Any], T], type[T]] = _default_cast,
        default: T = ...,
    ) -> T:
        ...

    def __call__(
        self,
        name: str,
        cast: Union[Callable[[Any], T], type[T]] = _default_cast,
        default: Union[T, type[MISSING]] = MISSING,
    ) -> T:
        return self.get(name, cast, default)



import importlib
import inspect
import os
import pathlib
import sys
import threading
import typing

from gyver.attrs import define

from gyver.exc import InvalidPath
from gyver.exc import MissingParams
from gyver.utils.lazy import lazyfield

PathConverter = typing.Callable[[pathlib.Path], str]
StrOrPath = typing.Union[str, pathlib.Path]
T = typing.TypeVar("T")
Validator = typing.Callable[[typing.Any], bool]


def make_modulename_resolver(
    root: pathlib.Path, path_converter: PathConverter = pathlib.Path.as_posix
) -> PathConverter:
    root_str = path_converter(root)

    def module_resolver(path: pathlib.Path) -> str:
        target_str = path_converter(path)
        return (
            target_str.replace(root_str, root.name)
            .removesuffix(".py")
            .replace("/", ".")
        )

    return module_resolver


def iterate_module(modpath: pathlib.Path, resolver: PathConverter):
    mod = importlib.import_module(resolver(modpath))
    yield from inspect.getmembers(mod)


@define
class _Finder:
    validator: Validator
    root: pathlib.Path
    look_on: typing.Optional[pathlib.Path] = None
    exclude: typing.Sequence[StrOrPath] = ()

    def __iter__(self):
        return self._iterdir(self.root)

    def find(self):
        return self.output

    @lazyfield
    def output(self):
        return dict(self)

    def __attrs_post_init__(self):
        if not self.root.exists():
            raise InvalidPath(f"root must be a valid path, received {self.root}")

    def _should_look(self, path: pathlib.Path):
        str_exclude = tuple(item for item in self.exclude if isinstance(item, str))
        path_exclude = tuple(
            item for item in self.exclude if isinstance(item, pathlib.Path)
        )
        return (
            path.name not in str_exclude
            and path not in path_exclude
            and (not path.is_file() or path.name.endswith(".py"))
        )

    @lazyfield
    def target_path(self):
        return self.look_on or self.root

    @lazyfield
    def resolver(self) -> PathConverter:
        return make_modulename_resolver(self.root)

    def _itermod(self, path: pathlib.Path):
        yield from (
            (name, obj)
            for name, obj in iterate_module(path, self.resolver)
            if self.validator(obj)
        )

    def _iterdir(
        self, path: pathlib.Path
    ) -> typing.Generator[tuple[str, typing.Any], None, None]:
        for item in path.iterdir():
            if not self._should_look(item):
                continue
            if item.is_dir():
                yield from self._iterdir(item)
            else:
                yield from self._itermod(item)


def class_validator(*types: type):
    if not types:
        raise MissingParams("Missing types in class_validator paramaters")

    def validator(obj: typing.Any):
        return (
            issubclass(obj, types) and obj not in types
            if inspect.isclass(obj)
            else False
        )

    return validator


def instance_validator(*types: type):
    if not types:
        raise MissingParams("Missing types in instance_validator paramaters")

    def validator(obj: typing.Any):
        return isinstance(obj, types)

    return validator


class FinderBuilder:
    def __init__(self) -> None:
        self.root: typing.Optional[pathlib.Path] = None
        self._validators = []
        self._lock = threading.Lock()
        self._exclude = []

    def reset(self):
        with self._lock:
            self.root = None
            self._validators = []
            self._exclude = []

    def from_path(self, path: pathlib.Path):
        if not path.exists():
            raise InvalidPath(f"root must be a valid path, received {path}")
        self.root = path
        return self

    def from_cwd(self):
        return self.from_path(pathlib.Path.cwd())

    def from_project(self):
        """
        Finds the root path of the project based on the file where this
        method was called. It works by checking the `sys.path` to find
        the longest path that contains the path to the file.
        This path is considered to be the root path of the project.

        Returns:
            The `FinderBuilder` object with the project root path set.
        """
        frame_stack: list[inspect.FrameInfo] = inspect.stack()

        frame_info: inspect.FrameInfo = frame_stack[1]

        if frame_info.filename == __file__:
            frame_info = next(
                frame for frame in frame_stack if frame.filename != __file__
            )

        caller_path: str = frame_info.filename
        caller_absolute_path: str = os.path.abspath(caller_path)
        caller_root_path = next(
            iter(
                sorted(
                    (p for p in sys.path if p in caller_absolute_path),
                    key=lambda p: len(p),
                )
            )
        )

        if not os.path.isabs(caller_path):
            caller_module_name: str = pathlib.Path(caller_path).name
            project_related_folders: str = caller_path.replace(
                os.sep + caller_module_name, ""
            )

            caller_root_path = caller_root_path.replace(project_related_folders, "")

        return self.from_path(pathlib.Path(caller_root_path))

    def from_package(self, stack_position: int = 1):
        """
        Finds the root path of the package that called the method and sets it as the
        base path for the finder.

        Parameters:
        stack_position (int, optional): The position in the stack trace of the calling
        function. The default is 1, which is the immediate caller.

        Returns:
        FinderBuilder: The FinderBuilder object, with the base path set to the
        root path of the calling package.
        """
        frame_stack = inspect.stack()

        frame_info = frame_stack[stack_position]

        if frame_info.filename == __file__:
            frame_info = next(
                frame for frame in frame_stack if frame.filename != __file__
            )

        package: str = frame_info.frame.f_globals["__name__"]
        steps_to_package = len(package.split("."))
        caller_root_path = pathlib.Path(package.replace(".", "/")).resolve()
        for _ in range(steps_to_package - 1):
            caller_root_path = caller_root_path.parent
        return self.from_path(pathlib.Path(caller_root_path))

    def _make_validator(self):
        validators = [*self._validators]

        def _validator(obj: typing.Any):
            return all(validator(obj) for validator in validators)

        return _validator

    def add_validator(self, validator: Validator):
        with self._lock:
            self._validators.append(validator)
        return self

    def child_of(self, *types: type):
        return self.add_validator(class_validator(*types))

    def instance_of(self, *types: type):
        return self.add_validator(instance_validator(*types))

    def exclude(self, *paths: StrOrPath):
        self._exclude.extend(paths)
        return self

    def exclude_default(self):
        return self.exclude("__init__.py", "__pycache__")

    def build(
        self,
        look_on: typing.Optional[pathlib.Path] = None,
    ) -> _Finder:
        with self._lock:
            if self.root is None:
                self.from_package()
            finder = _Finder(
                self._make_validator(),
                typing.cast(pathlib.Path, self.root),
                look_on,
            )
        self.reset()
        return finder

    def find(
        self,
        look_on: typing.Optional[pathlib.Path] = None,
    ):
        if self.root is None:
            self.from_project()
            assert self.root is not None
        return self.build(look_on).find()


finder_builder = FinderBuilder()



import threading
from queue import Empty
from queue import LifoQueue
from queue import Queue
from time import time
from typing import Callable
from typing import Generic
from typing import Optional
from typing import TypeVar

from gyver.attrs import call_init
from gyver.attrs import mutable

from gyver.exc import ErrorGroup

from .resource import Resource

T = TypeVar("T")


FactoryType = Callable[[], T]
ReleaserType = Callable[[T], None]


@mutable
class ThreadPool(Generic[T]):
    resources: Queue[Resource[T]]
    factory: FactoryType[T]
    releaser: ReleaserType[T]
    _pool_recycle: float
    _pool_size: int
    _available: int
    _available_semaphore: threading.Lock

    def __init__(
        self,
        factory: FactoryType[T],
        releaser: ReleaserType[T],
        queue_class: type[Queue] = LifoQueue,
        pool_size: int = 10,
        pool_recycle: float = 3600,
    ):
        call_init(
            self,
            resources=queue_class(pool_size),
            factory=factory,
            releaser=releaser,
            pool_recycle=pool_recycle,
            pool_size=pool_size,
            available=pool_size,
            available_semaphore=threading.Lock(),
        )

    def _initialize_resource(self) -> Resource[T]:
        resource = self.factory()
        return Resource.from_now(resource)

    def _maybe_recycle(
        self, resource: Resource[T], current: Optional[float] = None
    ) -> T:
        current = current or time()
        if (
            self._pool_recycle >= 0
            and resource.last_usage + self._pool_recycle <= current
        ):
            self.releaser(resource.get())
            resource = self._initialize_resource()
        return resource.get()

    def acquire(self) -> T:
        if self._decrease_available():
            try:
                resource = self.resources.get_nowait()
            except Empty:
                resource = self._initialize_resource()
        else:
            resource = self.resources.get()
        return self._maybe_recycle(resource)

    def release(self, resource: T) -> None:
        self.resources.put_nowait(Resource.from_resource(resource))
        self._increase_available()

    def prefill(self, count: Optional[int] = None) -> None:
        count = count or self._pool_size
        count = min(count, self._pool_size)

        while self.resources.qsize() < count:
            resource = self._initialize_resource()
            self.resources.put_nowait(resource)
            self._decrease_available()

    def dispose(self) -> None:
        errors = []
        while not self.resources.empty():
            resource = self.resources.get()
            try:
                self.releaser(resource.get())
            except Exception as e:
                errors.append(e)
            self._decrease_available()
        if errors:
            raise ErrorGroup("Could not kill all resources", *errors)

    def _decrease_available(self):
        """
        Decrease the count of available resources.

        This method should not be called directly. It is used internally
        by the `acquire` and `dispose` methods to keep track of the number
        of available resources.
        """
        with self._available_semaphore:
            if self._available == 0:
                return False
            self._available -= 1
            return True

    def _increase_available(self):
        """
        Increase the count of available resources.

        This method should not be called directly. It is used internally by the
        `release` and `prefill` methods to keep track of the number
        of available resources.
        """
        with self._available_semaphore:
            if self._available == self._pool_size:
                return False
            self._available += 1
            return True



from .adapter import AdapterConfigFactory
from .adapter.mark import as_config
from .adapter.mark import mark
from .config import MISSING
from .config import Config
from .config import EnvMapping
from .envconfig import DotFile
from .envconfig import EnvConfig
from .interface import ConfigLike
from .lazy import LazyConfig
from .typedef import Env
from .utils import boolean_cast

__all__ = [
    "boolean_cast",
    "Config",
    "EnvMapping",
    "MISSING",
    "AdapterConfigFactory",
    "mark",
    "as_config",
    "EnvConfig",
    "Env",
    "DotFile",
    "ConfigLike",
    "LazyConfig",
]



from collections import defaultdict
from typing import Mapping
from typing import Optional
from urllib.parse import urlparse
from urllib.parse import urlunsplit

from gyver.attrs import mutable

from gyver.url.encode import Encodable
from gyver.url.fragment import Fragment
from gyver.url.netloc import Netloc
from gyver.url.path import Path
from gyver.url.query import Query


@mutable(eq=False)
class URL(Encodable):
    scheme: str
    netloc: Netloc
    path: Path
    query: Query
    fragment: Fragment

    def __init__(self, val: str) -> None:
        self.load(val)

    def load(self, val: str):
        """
        Parse the given URL string and populate the properties of the URL object.

        :param val: The URL string.
        """
        parsed = urlparse(val)
        (
            self.scheme,
            netloc,
            path,
            _,
            query,
            fragment,
        ) = parsed
        self.query = Query(query)
        self.path = Path(path)
        self.fragment = Fragment(fragment)
        self.netloc = Netloc(netloc)

    def encode(self, omit_empty_equal: bool = True):
        """
        Encode the URL object as a URL string.

        :param omit_empty_equal: Whether to omit empty values with an equal sign in the query string.
        :return: The encoded URL string.
        """
        resolved_query = (
            self.query.encode() if omit_empty_equal else self.query.omit_empty_equal()
        )
        return urlunsplit(
            (
                self.scheme,
                self.netloc.encode(),
                self.path.encode(),
                resolved_query,
                self.fragment.encode(),
            )
        )

    def add(
        self,
        queryasdict: Optional[Mapping[str, str]] = None,
        /,
        path: Optional[str] = None,
        query: Optional[Mapping[str, str]] = None,
        fragment: Optional[str] = None,
        netloc: Optional[str] = None,
        netloc_obj: Optional[Netloc] = None,
    ):
        """
        Add components to the URL.

        :param queryasdict: Dictionary-like object representing query parameters.
        :param path: Path string to add to the URL.
        :param query: Dictionary-like object representing additional query parameters.
        :param fragment: Fragment string to set for the URL.
        :param netloc: Netloc string to set for the URL.
        :param netloc_obj: Netloc object to merge with the existing netloc.
        :return: The updated URL object.
        """
        if queryasdict:
            self.query.add(queryasdict)
        if path:
            self.path.add(path)
        if query:
            self.query.add(query)
        if fragment:
            self.fragment.set(fragment)
        if netloc:
            self.netloc = self.netloc.merge(Netloc(netloc))
        if netloc_obj:
            self.netloc = self.netloc.merge(netloc_obj)
        return self

    def set(
        self,
        queryasdict: Optional[Mapping[str, str]] = None,
        /,
        path: Optional[str] = None,
        query: Optional[Mapping[str, str]] = None,
        fragment: Optional[str] = None,
        netloc: Optional[str] = None,
        netloc_obj: Optional[Netloc] = None,
    ):
        """
        Set components of the URL.

        :param queryasdict: Dictionary-like object representing query parameters.
        :param path: Path string to set for the URL.
        :param query: Dictionary-like object representing query parameters.
        :param fragment: Fragment string to set for the URL.
        :param netloc: Netloc string to set for the URL.
        :param netloc_obj: Netloc object to set as the netloc.
        :return: The updated URL object.
        """
        if queryasdict:
            self.query.set(queryasdict)
        if path:
            self.path.set(path)
        if query:
            self.query.set(query)
        if fragment:
            self.fragment.set(fragment)
        if netloc:
            self.netloc.load(netloc)
        if netloc_obj:
            self.netloc = netloc_obj
        return self

    def copy(self):
        """
        Create a copy of the URL object.

        :return: The copied URL object.
        """
        new_url = URL("")
        new_url.scheme = self.scheme
        new_url.fragment.fragment_str = self.fragment.fragment_str
        new_url.path.segments = self.path.segments.copy()
        new_url.query.params = defaultdict(
            list,
            ((key, list(value)) for key, value in self.query.params.items()),
        )
        new_url.add(netloc_obj=self.netloc)
        return new_url



from .config import CryptoConfig
from .fernet import FernetCryptoProvider
from .rsa import RSACryptoConfig
from .rsa import RSACryptoProvider

__all__ = [
    "CryptoConfig",
    "FernetCryptoProvider",
    "RSACryptoConfig",
    "RSACryptoProvider",
]



from gyver import utils
from gyver.config import as_config
from gyver.exc import InvalidField
from gyver.utils.exc import panic

from . import drivers
from .typedef import Driver


@as_config
class DatabaseConfig:
    driver: Driver
    host: str
    port: int = -1
    user: str = ""
    password: str = ""
    name: str = ""
    pool_size: int = 20
    pool_recycle: int = 3600
    max_overflow: int = 0
    autotransaction: bool = False

    @utils.lazyfield
    def effective_port(self) -> int:
        return self.port if self.port != -1 else self.dialect.default_port

    @utils.lazyfield
    def dialect(self) -> drivers.Dialect:
        if self.driver is Driver.CUSTOM and not self.dialect_overriden:
            raise panic(
                InvalidField,
                'Field driver of value "custom" missing custom dialect,'
                "dont forget to call .override_dialect(dialect)",
            )
        return drivers.resolve_driver(self.driver)

    @utils.lazyfield
    def dialect_overriden(self):
        return False

    def override_dialect(self, dialect: drivers.Dialect) -> None:
        DatabaseConfig.dialect.manual_set(self, dialect)
        DatabaseConfig.dialect_overriden.manual_set(self, True)



from contextlib import contextmanager
from pathlib import Path

from gyver.attrs import mutable

from gyver.exc import FailedFileOperation
from gyver.utils import lazyfield

from .interface import AbstractFileTree
from .typedef import File
from .typedef import Folder
from .virtual import VirtualFileTree


@mutable
class FileTree(AbstractFileTree[Path]):
    base_dir: Path

    @lazyfield
    def root(self):
        return Folder.new(self.base_dir.name)

    @lazyfield
    def virtual(self):
        return VirtualFileTree(self.root)

    def write(self):
        if not self.base_dir.exists():
            self.base_dir.mkdir()
        self.write_folder(self.base_dir, self.root)

    @contextmanager
    def context(self):
        try:
            yield self
        except Exception as e:
            raise FailedFileOperation(
                "unable to write to disk after exception", e
            ) from e
        else:
            self.write()

    def write_file(self, path: Path, file: File):
        with open(path, "wb") as stream:
            stream.write(file.contents.getvalue())

    def write_folder(self, path: Path, folder: Folder):
        if not path.exists():
            path.mkdir()
        for value in folder.contents.values():
            inner_path = path / value.name
            if isinstance(value, File):
                self.write_file(inner_path, value)
            elif isinstance(value, Folder):
                self.write_folder(inner_path, value)

    def from_virtual(self, virtual_filetree: VirtualFileTree, *path: str):
        return self.virtual.from_virtual(virtual_filetree, *path)

    def virtual_context(self, dirname: str, *path: str):
        return self.virtual.virtual_context(dirname, *path)



from urllib.parse import quote

from gyver.attrs import mutable

from gyver.url.encode import Encodable
from gyver.url.utils import is_valid_encoded_path
from gyver.url.utils import utf8


@mutable(eq=False)
class Fragment(Encodable):
    """Represents an encodable URL fragment."""

    fragment_str: str

    def __init__(self, fragment_str: str) -> None:
        """
        Initialize the Fragment object.

        :param fragment_str: The fragment string.
        """
        self.set(fragment_str)

    def encode(self):
        """
        Encode the fragment string.

        :return: The encoded fragment string.
        """
        return self.fragment_str

    def set(self, fragment_str: str):
        """
        Set a new fragment string.

        If the provided string is not already encoded, it will be encoded.

        :param fragment_str: The new fragment string.
        :return: self
        """
        if not is_valid_encoded_path.match(fragment_str):
            fragment_str = quote(utf8(fragment_str))
        self.fragment_str = fragment_str
        return self

    def __str__(self):
        """
        Return the fragment string when converting the object to a string.

        :return: The fragment string.
        """
        return self.fragment_str



import inspect
import sys
from types import ModuleType
from typing import Any
from typing import Callable
from typing import Generic
from typing import Sequence
from typing import TypeVar
from typing import Union
from typing import overload

from gyver.attrs import define

from gyver.utils.lazy import lazyfield

from .interface import MISSING
from .interface import ConfigLike
from .interface import _default_cast

ConfigT = TypeVar("ConfigT", bound=ConfigLike)


@define
class EnvParam:
    """
    Represents an environment parameter.

    :param name: The name of the parameter.
    :param cast: A callable that casts the parameter value to the desired type.
    :param default: The default value of the parameter.
    :param frame_info: Frame information indicating where the parameter is defined.
    """

    name: str
    cast: Callable[[str], Any]
    default: Any
    frame_info: inspect.FrameInfo


T = TypeVar("T")


class LazyVal:
    def __init__(self, env_param: EnvParam, lazy_cfg: "LazyConfig") -> None:
        """
        Represents a lazily evaluated configuration value.

        :param env_param: The associated environment parameter.
        :param lazy_cfg: The parent LazyConfig object.
        """
        self.__env_param__ = env_param
        self.__lazy_cfg__ = lazy_cfg

    def __getattribute__(self, name: str) -> Any:
        """
        Retrieves the attribute value.

        :param name: The name of the attribute.
        :return: The attribute value.
        """
        if name in {"__env_param__", "__lazy_cfg__"}:
            return object.__getattribute__(self, name)
        resolved = self.__lazy_cfg__.resolve_param(self)
        return object.__getattribute__(resolved, name)


@define
class LazyConfig(Generic[ConfigT]):
    """
    Represents a lazily evaluated configuration.

    :param config: The configuration object.
    """

    config: ConfigT

    @lazyfield
    def inner_vals(self) -> list[LazyVal]:
        """
        Lazily evaluated values.

        :return: The list of lazily evaluated values.
        """
        return []

    def _get_stack(self):
        """
        Returns the next frame in the stack that is not from the current file.

        :return: The frame information.
        """
        return next(item for item in inspect.stack() if item.filename != __file__)

    def get(
        self,
        name: str,
        cast: Callable = _default_cast,
        default: Union[Any, type[MISSING]] = MISSING,
    ) -> Any:
        """
        Retrieves the lazily evaluated value.

        :param name: The name of the value.
        :param cast: A callable that casts the value to the desired type.
        :param default: The default value if the parameter is missing.
        :return: The lazily evaluated value.
        """
        lazyval = LazyVal(EnvParam(name, cast, default, self._get_stack()), self)
        self.inner_vals.append(lazyval)
        return lazyval

    @overload
    def __call__(
        self,
        name: str,
        cast: Union[Callable[[Any], T], type[T]] = _default_cast,
        default: type[MISSING] = MISSING,
    ) -> T:
        ...

    @overload
    def __call__(
        self,
        name: str,
        cast: Union[Callable[[Any], T], type[T]] = _default_cast,
        default: T = ...,
    ) -> T:
        ...

    def __call__(
        self,
        name: str,
        cast: Union[Callable[[Any], T], type[T]] = _default_cast,
        default: Union[T, type[MISSING]] = MISSING,
    ) -> T:
        """
        Retrieves the lazily evaluated value.

        :param name: The name of the value.
        :param cast: A callable that casts the value to the desired type.
        :param default: The default value if the parameter is missing.
        :return: The lazily evaluated value.
        """
        return self.get(name, cast, default)

    @lazyfield
    def modcache(self) -> dict[ModuleType, Sequence[tuple[str, Any]]]:
        """
        Cached module members.

        :return: The dictionary mapping modules to their members.
        """
        return {}

    def _get_members(self, mod: ModuleType) -> Sequence[tuple[str, Any]]:
        """
        Retrieves the members of the given module.

        :param mod: The module object.
        :return: The sequence of module members.
        """
        if mod in self.modcache:
            return self.modcache[mod]
        self.modcache[mod] = inspect.getmembers(mod)
        return self.modcache[mod]

    def resolve_param(self, lazyval: "LazyVal"):
        """
        Resolves the lazily evaluated parameter.

        :param lazyval: The LazyVal object.
        :return: The resolved value.
        """
        envparam = lazyval.__env_param__
        mod = sys.modules[envparam.frame_info.frame.f_globals["__name__"]]
        resolved = self.config.get(envparam.name, envparam.cast, envparam.default)
        for key, value in self._get_members(mod):
            if getattr(value, "__env_param__", None) is envparam:
                setattr(mod, key, resolved)
        return resolved

    def resolve(self):
        """
        Resolves all the lazily evaluated parameters.
        """
        for val in self.inner_vals:
            self.resolve_param(val)
        self.modcache.clear()

    def expected(self) -> list[EnvParam]:
        """
        Retrieves the list of expected environment parameters.

        :return: The list of expected environment parameters.
        """
        return [item.__env_param__ for item in self.inner_vals]



from . import query
from .adapter import DatabaseAdapter
from .config import DatabaseConfig
from .context import AsyncSaContext
from .context import AsyncSessionContext
from .context import SaContext
from .context import SessionContext
from .drivers import Dialect
from .entity import AbstractEntity
from .entity import Entity
from .entity import make_table
from .metadata import metadata as default_metadata
from .typedef import Driver
from .utils import make_uri

__all__ = [
    "Dialect",
    "DatabaseConfig",
    "Driver",
    "make_uri",
    "DatabaseAdapter",
    "SaContext",
    "AsyncSaContext",
    "SessionContext",
    "AsyncSessionContext",
    "Entity",
    "AbstractEntity",
    "make_table",
    "query",
    "default_metadata",
]



def boolean_cast(string: str):
    """
    The boolean_cast function converts a string to its boolean equivalent.
    1 and true(case-insensitive) are considered True, everything else, False.

    :param string: str: Check if the string is true or false
    :return: A boolean value based on the string it receives
    """
    return {
        "true": True,
        "false": False,
        "1": True,
        "0": False,
        "": False,
    }.get(string.lower(), False)



from typing import Generic

from gyver.context.context import AsyncContext
from gyver.context.context import Context
from gyver.context.interfaces.adapter import AtomicAdapter
from gyver.context.interfaces.adapter import AtomicAsyncAdapter
from gyver.context.typedef import T


class BoundContext(Context[T], Generic[T]):
    adapter: AtomicAdapter[T]

    def __init__(self, adapter: AtomicAdapter[T], context: Context[T]) -> None:
        super().__init__(adapter)
        self._context = context

    def acquire(self):
        with self._lock:
            client = self._context.acquire()
            self._stack += 1
            self.adapter.begin(client)
        return client

    def release(self, commit: bool = True):
        with self._lock:
            if self._stack == 1:
                if self.adapter.in_atomic(self._context.client):
                    if commit:
                        self.adapter.commit(self._context.client)
                    else:
                        self.adapter.rollback(self._context.client)
                self._context.release()
            self._stack -= 1

    def __exit__(self, *exc):
        self.release(not any(exc))


class AsyncBoundContext(AsyncContext[T], Generic[T]):
    adapter: AtomicAsyncAdapter[T]

    def __init__(
        self, adapter: AtomicAsyncAdapter[T], context: AsyncContext[T]
    ) -> None:
        super().__init__(adapter)
        self._context = context

    async def acquire(self):
        async with self._lock:
            client = await self._context.acquire()
            if self.stack == 0:
                await self.adapter.begin(client)
            self._stack += 1
            return client

    async def release(self, commit: bool = True):
        async with self._lock:
            if self._context._client is None:
                raise RuntimeError(
                    "Release should not be called before client initialization"
                )
            if self._stack == 1:
                if commit:
                    await self.adapter.commit(self._context._client)
                else:
                    await self.adapter.rollback(self._context._client)
            self._stack -= 1
        await self._context.release()

    async def __aexit__(self, *exc):
        await self.release(not any(exc))



from enum import Enum
from typing import NamedTuple

from typing_extensions import Self


class EnvTuple(NamedTuple):
    val: str
    weight: int


class Env(EnvTuple, Enum):
    LOCAL = EnvTuple("local", 1)
    TEST = EnvTuple("test", 2)
    DEV = EnvTuple("dev", 3)
    PRD = EnvTuple("prd", 4)

    @property
    def value(self) -> EnvTuple:
        return super().value

    @property
    def val(self):
        return self.value.val

    @property
    def weight(self):
        return self.value.weight

    @classmethod
    def new(cls, val: str) -> Self:
        return next(value for value in cls if value.val == val)



from typing import Generic

from gyver.context.context import AsyncContext
from gyver.context.context import Context
from gyver.context.interfaces.adapter import AtomicAdapter
from gyver.context.interfaces.adapter import AtomicAsyncAdapter
from gyver.context.typedef import T
from gyver.utils import lazyfield


class AtomicContext(Context[T], Generic[T]):
    adapter: AtomicAdapter[T]

    def __init__(self, adapter: AtomicAdapter[T]) -> None:
        super().__init__(adapter)

    @lazyfield
    def client(self):
        return self.adapter.new()

    def acquire(self):
        with self._lock:
            if self.adapter.is_closed(self.client):
                AtomicContext.client.cleanup(self)
            self._stack += 1
            self.adapter.begin(self.client)
        return self.client

    def release(self, commit: bool = True):
        with self._lock:
            if self._stack == 1:
                if self.adapter.in_atomic(self.client):
                    if commit:
                        self.adapter.commit(self.client)
                    else:
                        self.adapter.rollback(self.client)
                self.adapter.release(self.client)
                Context.client.cleanup(self)
            self._stack -= 1

    def __exit__(self, *exc):
        self.release(not any(exc))


class AsyncAtomicContext(AsyncContext[T], Generic[T]):
    adapter: AtomicAsyncAdapter[T]

    def __init__(self, adapter: AtomicAsyncAdapter[T]) -> None:
        super().__init__(adapter)

    async def acquire(self):
        async with self._lock:
            client = await self.client()
            if self.stack == 0:
                await self.adapter.begin(client)
            self._stack += 1
            return client

    async def release(self, commit: bool = True):
        async with self._lock:
            if self._client is None:
                raise RuntimeError(
                    "Release should not be called before client initialization"
                )
            if self._stack == 1:
                if await self.adapter.in_atomic(self._client):
                    if commit:
                        await self.adapter.commit(self._client)
                    else:
                        await self.adapter.rollback(self._client)
                await self.adapter.release(self._client)
                self._client = None
            self._stack -= 1

    async def __aexit__(self, *exc):
        await self.release(not any(exc))



import os
from pathlib import Path
from typing import Any
from typing import Callable
from typing import Sequence
from typing import Union

from gyver.attrs import call_init
from gyver.attrs import define
from gyver.attrs import info

from gyver import utils
from gyver.config.config import MISSING

from .config import Config
from .config import EnvMapping
from .config import default_mapping
from .typedef import Env


@define
class DotFile:
    filename: Union[str, Path] = info(order=False)
    env: Env = info(order=lambda env: env.weight)
    apply_to_lower: bool = info(default=False, order=False)

    def is_higher(self, env: Env) -> bool:
        return self.env.weight >= env.weight


def default_rule(_: Env):
    return False


@define
class EnvConfig(Config):
    mapping: EnvMapping = default_mapping
    env_var: str = "CONFIG_ENV"
    env_cast: Callable[[str], Env] = Env
    dotfiles: Sequence[DotFile] = ()
    ignore_default_rule: Callable[[Env], bool] = default_rule

    def __init__(
        self,
        *dotfiles: DotFile,
        env_var: str = "CONFIG_ENV",
        mapping: EnvMapping = default_mapping,
        ignore_default_rule: Callable[[Env], bool] = default_rule,
        env_cast: Callable[[str], Env] = Env
    ) -> None:
        call_init(
            self,
            env_var=env_var,
            mapping=mapping,
            dotfiles=dotfiles,
            ignore_default_rule=ignore_default_rule,
            env_cast=env_cast,
        )

    def __post_init__(self):
        if self.dotfile:
            EnvConfig.file_values.manual_set(
                self, dict(self._read_file(self.dotfile.filename))
            )

    @utils.lazyfield
    def env(self):
        return Config.get(self, self.env_var, Env.new)

    @utils.lazyfield
    def ignore_default(self):
        return self.ignore_default_rule(self.env)

    def get(
        self,
        name: str,
        cast: Callable[..., Any] = ...,
        default: Union[Any, type[MISSING]] = ...,
    ) -> Any:
        default = MISSING if self.ignore_default else default
        return Config.get(self, name, cast, default)

    @utils.lazyfield
    def dotfile(self):
        """
        The dotfile function returns the dotfile object that is applicable to
        the current environment.  It traverses the list of dotfiles in reverse order,
        so that higher priority dotfiles are checked first.  If no matching dotfile is
        found, it returns None.

        :return: The first dotfile that is not higher than the current environment
        """

        for dot in sorted(self.dotfiles, reverse=True):
            if not dot.is_higher(self.env):
                break
            if dot.env != self.env and (
                not dot.apply_to_lower or not dot.is_higher(self.env)
            ):
                continue
            if not os.path.isfile(dot.filename):
                continue
            return dot



import typing

from typing_extensions import Self

T = typing.TypeVar("T")
SelfT = typing.TypeVar("SelfT")

_obj_setattr = object.__setattr__


class lazyfield(typing.Generic[SelfT, T]):
    """
    A descriptor class that can be used as a decorator for a method on a class.
    When the decorated method is accessed on an instance, it will check if the
    instance has an attribute with the same name as the method but with an
    underscore prefix. If the attribute does not exist, it will call the decorated
    method on the instance and set the result as the attribute's value.
    Subsequent accesses will return the cached value, avoiding unnecessary
    recalculation or computation.
    """

    def __init__(self, func: typing.Callable[[SelfT], T]) -> None:
        """
        func : callable
            The function that will be decorated. This function should take
            a single argument, which is the instance of the class it is a
            method of.
        """
        self._func = func

    def __set_name__(self, owner: type[SelfT], name: str):
        self.public_name = name
        self.private_name = f"_lazyfield_{name}"

    @typing.overload
    def __get__(self, instance: SelfT, owner: type[SelfT]) -> T:
        ...

    @typing.overload
    def __get__(self, instance: typing.Literal[None], owner: type[SelfT]) -> Self:
        ...

    def __get__(
        self,
        instance: typing.Optional[SelfT],
        owner: typing.Optional[type[SelfT]] = None,
    ) -> typing.Union[T, Self]:
        if not instance:
            return self
        try:
            val = typing.cast(
                T,
                object.__getattribute__(
                    instance,
                    self.private_name,
                ),
            )
        except AttributeError:
            val = self._try_set(instance)
        return val

    def _try_set(self, instance: SelfT) -> T:
        try:
            val = self._func(instance)
        except Exception as e:
            # remove exception context to create easier traceback
            raise e from None
        else:
            _obj_setattr(instance, self.private_name, val)
            return val

    def __set__(self, instance: SelfT, value: T):
        self.manual_set(instance, value)

    def __delete__(self, instance: SelfT):
        self.cleanup(instance)

    def cleanup(self, instance: SelfT):
        object.__delattr__(instance, self.private_name)

    def manual_set(self, instance: SelfT, value: T):
        _obj_setattr(instance, self.private_name, value)



import sys
from itertools import chain
from typing import Sequence


class GyverError(Exception):
    """Base exception for all gyver exceptions"""


class InvalidCast(GyverError):
    """Exception raised for config when
    cast callable raises an error"""


class MissingName(GyverError, KeyError):
    """Exception raised for config when
    name is not found in given environ"""


class InvalidField(GyverError, KeyError):
    """Exception raised when field lookup
    returned an unexpected result on mapping"""


class CacheMiss(GyverError, KeyError):
    """Exception raised when cache lookup
    returns None or raises error"""


class QueueNotFound(InvalidField):
    """Exception raised when queue name provider
    does not exist on given credentials"""


class MissingParams(GyverError, NotImplementedError):
    """Exception raised when no params where passed in a
    *args function
    """


class InvalidPath(GyverError, ValueError):
    """Equivalent to FileNotFound but also for directories"""


class InvalidParamType(GyverError, TypeError):
    """Exception raised when parameter received
    is not of the type excpected"""


class InvalidParamValue(GyverError, ValueError):
    """Exception raised when parameter received
    has an unexpected value"""


if sys.version_info < (3, 11):

    class ErrorGroup(GyverError):
        exceptions: tuple[Exception, ...]

        def __init__(self, message: str, exceptions: Sequence[Exception]) -> None:
            self.message = message
            self.exceptions = tuple(exceptions)
            args = list(chain(item.args for item in exceptions))
            super().__init__(self.message, *args)

        def __str__(self) -> str:
            suffix = "" if len(self.exceptions) == 1 else "s"
            return f"{self.message} ({len(self.exceptions)} sub-exception{suffix})"

        def __repr__(self) -> str:
            return f"{self.__class__.__name__}({self.message!r}, {self.exceptions!r})"

else:

    class ErrorGroup(GyverError, ExceptionGroup): # noqa: F821
        pass


class FailedFileOperation(GyverError):
    """Exception raised when an error happened while interacting with a file"""


class MergeConflict(GyverError):
    """Exception raised when merge_dicts(strict=True) finds a conflict"""



import abc
from typing import Union


class Encodable(abc.ABC):
    @abc.abstractmethod
    def encode(self) -> str:
        raise NotImplementedError

    def __eq__(self, other: Union["Encodable", str]) -> bool:
        if isinstance(other, str):
            return self.encode() == other
        return self.encode() == other.encode() or super().__eq__(other)

    def __str__(self) -> str:
        return self.encode()



from typing import Any
from typing import Generator
from typing import Sequence
from typing import Union

from gyver.attrs import define
from gyver.attrs import fields
from gyver.attrs.field import Field
from gyver.attrs.utils.typedef import MISSING as NOTHING

from gyver.config.adapter.interface import FieldResolverStrategy
from gyver.config.config import MISSING


@define
class GyverAttrsResolverStrategy(FieldResolverStrategy[Field]):
    field: Field

    def cast(self) -> type:
        return self.field.declared_type

    def names(self) -> Sequence[str]:
        return self.field.name, self.field.alias

    def init_name(self) -> str:
        return self.field.alias or self.field.name

    def default(self) -> Union[Any, type[MISSING]]:
        default = self.field.default
        if default is NOTHING:
            return MISSING
        return default() if self.field.has_default_factory else default

    @staticmethod
    def iterfield(config_class: type) -> Generator[Field, Any, Any]:
        yield from fields(config_class).values()



import sqlalchemy as sa

from gyver.database.typedef import ClauseType

from . import interface


class NullBind(interface.BindClause):
    type_ = ClauseType.BIND

    def bind(self, mapper: interface.Mapper) -> interface.Comparison:
        del mapper
        return sa.true()



from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.asymmetric import rsa
from gyver.attrs import define
from gyver.attrs import info

from gyver.config import AdapterConfigFactory
from gyver.config import as_config
from gyver.utils import lazyfield


@as_config
class RSACryptoConfig:
    __prefix__ = "crypto"

    private_key: str
    public_key: str


@define
class RSACryptoProvider:
    config: RSACryptoConfig = info(
        default_factory=AdapterConfigFactory().maker(RSACryptoConfig)
    )

    @lazyfield
    def _private_key(self) -> rsa.RSAPrivateKey:
        return serialization.load_pem_private_key(  # type: ignore
            self.config.private_key.encode(),
            password=None,
            backend=default_backend(),
        )

    @lazyfield
    def _public_key(self) -> rsa.RSAPublicKey:
        return serialization.load_pem_public_key(  # type:ignore
            self.config.public_key.encode(), backend=default_backend()
        )

    def encrypt(self, secret: str) -> str:
        ciphertext = self._public_key.encrypt(
            secret.encode(),
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None,
            ),
        )
        return ciphertext.hex()

    def decrypt(self, ciphertext: str) -> str:
        encoded = bytes.fromhex(ciphertext)
        return self._private_key.decrypt(
            encoded,
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None,
            ),
        ).decode()



import functools
import warnings
from collections import deque
from itertools import chain
from typing import Callable
from typing import Literal
from typing import TypeVar
from typing import cast

from gyver.attrs import define
from typing_extensions import ParamSpec

from gyver.exc import MergeConflict

from .exc import panic


def _not_implemented(msg: str):
    def inner(*_, **__kwds__):
        raise panic(NotImplementedError, msg)

    return inner


P = ParamSpec("P")
T = TypeVar("T")


def cache(f: Callable[P, T]) -> Callable[P, T]:
    return cast(Callable[P, T], functools.cache(f))


def deprecated(func: Callable[P, T]) -> Callable[P, T]:
    @functools.wraps(func)
    def inner(*args: P.args, **kwargs: P.kwargs) -> T:
        if not hasattr(func, "__warn_deprecated__"):
            warnings.warn(
                f"{func.__qualname__} is "
                "deprecated and can be removed without notice",
                DeprecationWarning,
            )
            func.__warn_deprecated__ = True
        return func(*args, **kwargs)

    return inner


frozen = deprecated(define)


class DeprecatedClass:
    __warn_deprecated__ = False

    def __init__(self) -> None:
        if not type(self).__warn_deprecated__:
            warnings.warn(
                f"{'.'.join((type(self).__module__, type(self).__name__))} is "
                "deprecated and can be removed without notice",
                DeprecationWarning,
            )
            type(self).__warn_deprecated__ = True


def merge_dicts(
    left: dict,
    right: dict,
    on_conflict: Literal["strict", "left", "right"],
    merge_sequences: bool = True,
) -> dict:
    """
    Merge two dictionaries with customizable conflict resolution strategy.

    :param left: The left dictionary to merge.
    :type left: dict
    :param right: The right dictionary to merge.
    :type right: dict
    :param on_conflict: The conflict resolution strategy to use.
        - 'strict': Raise a MergeConflict exception if conflicts occur.
        - 'left': Prioritize the values from the left dictionary in case of conflicts.
        - 'right': Prioritize the values from the right dictionary in case of conflicts.

    :type on_conflict: Literal["strict", "left", "right"]
    :param merge_sequences: Indicates whether to merge sequences (lists, sets, tuples) or skip them.
        If True, sequences will be merged based on the conflict resolution strategy.
        If False, sequences will be skipped, and the value from the chosen (defaults to left on strict)
            dictionary will be used.
    :type merge_sequences: bool, optional
    :return: The merged dictionary.
    :rtype: dict

    :raises MergeConflict: If conflicts occur and the conflict resolution strategy is set to 'strict'.
    """

    output = {key: value for key, value in left.items() if key not in right}

    stack = deque([(left, right, output)])

    while stack:
        left_curr, right_curr, output_curr = stack.pop()

        for key, value in right_curr.items():
            if key not in left_curr:
                output_curr[key] = value
            elif isinstance(value, (list, set, tuple)) and merge_sequences:
                left_val = left_curr[key]
                if isinstance(left_val, (list, set, tuple)):
                    type_ = type(value) if on_conflict == "right" else type(left_val)
                    output_curr[key] = type_(chain(left_val, value))
            elif isinstance(value, dict):
                if isinstance(left_curr[key], dict):
                    output_curr[key] = {
                        lkey: lvalue
                        for lkey, lvalue in left_curr[key].items()
                        if lkey not in value
                    }
                    stack.append((left_curr[key], value, output_curr[key]))
            elif on_conflict not in ("left", "right"):
                raise MergeConflict(
                    "Conflict found when trying to merge dicts",
                    key,
                    value,
                    left_curr[key],
                )
            elif on_conflict == "left":
                output_curr[key] = left_curr[key]
            else:
                output_curr[key] = value

    return output



import re
import shlex
from typing import Callable
from typing import TypeVar

_to_camel_regexp = re.compile("_([a-zA-Z])")
_to_snake_regexp = re.compile("([a-z])([A-Z])")


def to_snake(camel_string: str) -> str:
    """
    The to_snake function converts camelCase strings to snake_case.

    :param camel_string: str: Pass the camelcase string to be converted
    :return: A string in snake_case
    """
    return _to_snake_regexp.sub(r"\1_\2", camel_string).lower()


def to_camel(snake_string: str) -> str:
    """
    The to_camel function converts a snake_string to camelCase.

    :param snake_string: str: Pass the string that is to be converted
    :return: A string that is converted from snake_case to camelcase
    """
    return _to_camel_regexp.sub(
        lambda match: match[1].upper(), snake_string
    ).removesuffix("_")


def to_pascal(snake_string: str):
    """
    The to_pascal function converts a snake_case string to an PascalCase string.

    :param snake_string: str: Pass in the string that we want to convert
    :return: A string that is the snake_string converted to camel
    case and with the first letter capitalized
    """
    val = to_camel(snake_string)
    return val and val[0].upper() + val[1:]


# Compat
upper_camel = to_pascal


def to_lower_kebab(camel_string: str) -> str:
    """
    The to_lower_kebab function converts camelCase strings to kebab-case.

    :param camel_string: str: Pass the camelCase string to be converted
    :return: A string in kebab-case
    """
    snake_string = to_snake(camel_string)
    return snake_string.replace("_", "-")


T = TypeVar("T")

OuterCastT = TypeVar("OuterCastT", list, tuple, set)


def make_lex_separator(
    outer_cast: type[OuterCastT], cast: type = str
) -> Callable[[str], OuterCastT]:
    """
    Create a lexer-based separator function.

    :param outer_cast: The type to cast the resulting separated values into (e.g., list, tuple, set).
    :param cast: The type to which each individual item will be cast (default: str).
    :return: A callable that separates a string into an instance of outer_cast with casted items.

    Usage example:
    comma_separated = make_lex_separator(tuple, str)
    result = comma_separated("a, b, c")
    print(result)  # Output: ('a', 'b', 'c')
    """

    def wrapper(value: str) -> OuterCastT:
        lex = shlex.shlex(value, posix=True)
        lex.whitespace = ","
        lex.whitespace_split = True
        return outer_cast(cast(item.strip()) for item in lex)

    return wrapper


comma_separated = make_lex_separator(tuple, str)
